#ifndef pef2D_H
#define pef2D_H

#include <stdexcept>
#include <cstdlib>
#include "float2DReg.h"
#include "hypercube.h"
#include "axis.h"
#include "floatHyperExt.h"
#include "cgls.h"

using namespace SEP;

//2D Prediction Error Filters
class pef2D {
protected:
    // leading 1 position on the first column
    // and size of gap following it if any
    unsigned int _lead, _gap;

public:
    // default constructor
    pef2D(){}

    // constructor
    pef2D(unsigned int lead=0, unsigned int gap=0){_lead = lead; _gap = gap;}

    // get the parameters
    unsigned int getLead(){return _lead;}

    unsigned int getGap(){return _gap;}

    // set the parameters
    void setLead(const unsigned int lead){_lead = lead;}

    void setGap(const unsigned int gap){_gap = gap;}

    // estimate stationary pef2D from data
    void estimate(std::shared_ptr<float2DReg> pef, const std::shared_ptr<float2DReg> dat, cgls &cg, const bool crossBounds = true);

};

//Non-stationary 2D PEF
class nonStatPef2D : public pef2D {
protected:
    // damping factor for step length calculation when updating the PEF
    // it is used as step lenght when the switch is ON
    double _epsilon;

    // weights for computing new PEF from x and z contributions
    double _weightx, _weightz;

    // damping factor for weights calculation when computing new PEF from x and z contributions
    double _lambda;

    // switch to new implementation (the original one is from J. Claerbout)
    bool _switch;

    // switch to force the weights to user defined values
    bool _forceWeights;

public:
    // constructor
    nonStatPef2D(unsigned int lead=0, unsigned int gap=0, double epsilon = 1e-01, double lambda = 1e+03) : pef2D(lead, gap){
        if (epsilon == 0){
            std::clog << "Epsilon must be non-zero. It will be reset to 0.1\n";
            epsilon = 0.1;
        }
        if (lambda == 0){
            std::clog << "Lambda must be non-zero. It will be reset to 1000\n";
            lambda = 1000;
        }
        _epsilon = epsilon;
        _lambda = lambda;
        _switch = 0;
        _weightx = 1;
        _weightz = 1;
        _forceWeights = 0;
    }

    // set the parameters
    void setEpsilon(const double epsilon){_epsilon = epsilon;}

    void setLambda(const double lambda){_lambda = lambda;}

    void setSwitch(const bool sw) {_switch = sw;}

    void forceWeights(double weightx = 1, double weightz = 1){
        if ((weightx == 0.0) && (weightz == 0.0))
            throw std::logic_error("At least one of the weights should be non-zero\n");

        _weightx = weightx;
        _weightz = weightz;
        _forceWeights = 1;
    }

    void unforceWeights(){_forceWeights = 0;}

    // estimate and apply the PEF to 2D data; the last PEF is stored and the residual is output
    void apply(std::shared_ptr<float2DReg> pef, const std::shared_ptr<float2DReg> dat, std::shared_ptr<float2DReg> res, const bool crossBounds = true);
    
};

#endif