#ifndef Trace_H
#define Trace_H

#include <iostream>
#include <stdio.h>
#include <cmath>
#include <math.h>
#include <fstream>
#include <string>
#include <cstring>
#include <vector>
#include <array>
#include <stdlib.h>
#include <ctime>
#include <complex>


#define TYPE double

using namespace std;


class Trace {
private:
	vector<TYPE> * vec; 	// data
	int t0;				// Origin time in terms of number of samples
	TYPE si;	// Sampling interval
public:
	// Default constructor
	Trace(){si=1; t0=0; vec = nullptr;}
	
	// Destructor
	~Trace(){delete vec;}
	
	// Constructor
	Trace(vector<TYPE> * v, TYPE SI=1, int t=0){
		t0 = t; si = SI ;
		vec = v;
	}
	
	// Constructor by copying
	Trace(const Trace &tr){
		t0 = tr.t0 ; si = tr.si ;
		vec = new vector<TYPE>();
		(*vec).assign((*tr.vec).begin(),(*tr.vec).end()); 
	}
	
	// Constructor from an ascii file with a special format
	Trace(string file){
		ifstream in; in.open(file); cout << "Open failed: "<< in.fail() << endl;
		string line;
		stringstream linestream;
		getline(in,line);
		line.erase(0,4); linestream << line; linestream >> si; linestream.str("");linestream.clear();
		getline(in,line);
		line.erase(0,4); linestream << line; linestream >> t0; linestream.str("");linestream.clear();
		TYPE value;
		vec = new vector<TYPE>();
		while (getline(in,line)){
			value = stoi(line);
			(*vec).push_back(value);
		}
		in.close();
	}

	// Overload the assignment operator
	void operator = (const Trace &tr){
		t0 = tr.t0 ; si = tr.si;
		delete vec;
		vec = new vector<TYPE>();
		(*vec).assign((*tr.vec).begin(),(*tr.vec).end());
	}

	// Get the trace SI
	TYPE get_si() {return si;}
	
	// Get the trace origine time
	int get_t0() {return t0;}
	
	// Get the trace vector pointer
	vector<TYPE> * get_vector(){
		return vec;
	}

	// Get the pointer to the vector data
	TYPE * get_data(){
		return (*vec).data();
	}

	// Get the trace size
	unsigned int size(){
		return (*vec).size();
	}
	
	// Set the trace SI
	void set_si(TYPE SI) {si = SI;}
	
	// Set the trace origine time
	void set_t0(int t) {t0 = t;}

	// Set the trace vector
	void set_vector(vector<TYPE> * v){
		delete vec;
		vec=v;}

	// Copy the vector data
	void copy_data(vector<TYPE> * v){
		delete vec;
		vec = new vector<TYPE>();
		(*vec).assign((*v).begin(),(*v).end());
	}
	
	// Display the values of the trace
	void print(string message="", ostream &o=cout);

	// Save trace to an ascii file with a special format
	void save(string file);
	
	// Compute norm2 squared of the trace
	TYPE norm2();
	
	// Shift a trace by a number of samples
	void sshift(int n=0){
		t0=t0+n;
	}
	
	// Resize (cut or pad at the end) the trace
	void resize(unsigned int n=0){
		vec->resize(n);
	}
	
	// Pad zeros at the beginning
	void pad_start(unsigned int n=0){
		t0 = t0 - n;
		for (int i=0; i<n; i++) {vec->insert(vec->begin(),0);}
	}

    // Erase elements from the beginning
    void erase_start(unsigned int n=0){
        t0 = t0 + n;
        vec->erase(vec->begin(), vec->begin()+n);
    }

	// Add a constant to the trace
	Trace operator + (const TYPE c);

	// Add two traces together
	Trace operator + (Trace &tr);

	// Subtract a constant from the trace
	Trace operator - (const TYPE c);

	// Subtract one trace from another
	Trace operator - (Trace &tr);

	// Multiply the trace by a constant
	Trace operator * (const TYPE c);

	// Divide the trace by a non-zero constant
	Trace operator / (const TYPE c);

	// Scale the trace by a constant
	void scale(const TYPE c);

	// Scale the trace by another trace
	void scale(Trace &tr);
	
    // Dot product between two traces
    TYPE operator * (Trace &f);

	// Differentiate the trace
	void differentiate(bool freq_domain=false);

	// Integrate the trace
	void integrate(bool freq_domain=false);

	// Interpolate data in time or frequency domain
	void interp(unsigned int factor=1, bool freq_domain=true);

	// Apply FIR Hanning filter
	void fir_hanning();

	// Apply FIR windowed sinc filter
	// cutoff is the percentage of Nyquist frequency
	// half_len is the half length of sinc operator, in number of samples
	// alpha is the coefficient used for cosine windowing
	void fir_sinc(TYPE cutoff, int half_len=13, TYPE alpha=0.5);

	// Apply IIR Butterworth filter
	// cutoff is the percentage of Nyquist frequency
	// half_order is the half order of the filter
	// visit http://www.kwon3d.com/theory/filtering/butt.html for equations
	// visit http://www.ee.ic.ac.uk/hp/staff/dmb/courses/DSPDF/00800_TransIIR.pdf
	// for filters transformation (slide 66)
	void iir_butterworth(TYPE cutoff, bool lowpass = true, unsigned int half_order=2);

	// Apply IIR Butterworth band-pass filter
	// cutoff is the percentage of Nyquist frequency
	// half_order is the half order of the filter
	void iir_butterworth_bp(TYPE cutoff_low, TYPE cutoff_high, unsigned int half_order=2);

	// Transform the trace into a zero-phase equivalent
	void zero_phase();	

};

// Convolve a trace with a filter
Trace* convolve(Trace* tr, Trace* f, bool truncate = false);

// Cross correlate a trace with a filter
Trace* correlate(Trace* tr, Trace* f, bool truncate = false);

#endif