#include <fstream>
#include <stdio.h>
#include "floatHyperExt.h"
#include "axis.h"
#include <stdexcept>
#include <cstdlib>
#include <math.h>
#include <algorithm>
#include "functions.h"
#include "hypercube.h"
#include "axis.h"
#include "fkTransform.h"


using namespace SEP;

float VECEXT::getox(const std::shared_ptr<float2DReg> vec) {
    return vec->getHyper()->getAxis(2).o;}

float VECEXT::getoz(const std::shared_ptr<float2DReg> vec) {
return vec->getHyper()->getAxis(1).o;}

float VECEXT::getdx(const std::shared_ptr<float2DReg> vec) {
return vec->getHyper()->getAxis(2).d;}

float VECEXT::getdz(const std::shared_ptr<float2DReg> vec) {
return vec->getHyper()->getAxis(1).d;}

unsigned int VECEXT::getnx(const std::shared_ptr<floatHyper> vec) {
return vec->getHyper()->getAxis(2).n;}

unsigned int VECEXT::getnz(const std::shared_ptr<floatHyper> vec) {
return vec->getHyper()->getAxis(1).n;}

void VECEXT::setox(std::shared_ptr<floatHyper> vec, float ox) {
    std::vector<axis> axes = vec->getHyper()->getAxes();
    axes[1].o = ox;
    std::shared_ptr<hypercube> h (new hypercube(axes));
    vec->setHyper(h);
}

void VECEXT::setoz(std::shared_ptr<floatHyper> vec, float oz) {
    std::vector<axis> axes = vec->getHyper()->getAxes();
    axes[0].o = oz;
    std::shared_ptr<hypercube> h (new hypercube(axes));
    vec->setHyper(h);
}

void VECEXT::write(const std::shared_ptr<floatHyper> vec, std::string name) {
    unsigned long long n123 = vec->getHyper()->getN123();
    std::ofstream output;
    output.open(name+".bin", std::ios::out | std::ios::binary);
    if (output.is_open())
	    output.write((char*)vec->getVals(), n123*sizeof(float));
	output.close();

    std::vector<axis> axes = vec->getHyper()->getAxes();
    output.open(name+".header");
    if (output.is_open()){
        output << "type=float"<< axes.size() << "DReg\n";
        for (int i=0; i<axes.size(); i++){
            output << "n"<< i+1 <<"=" << axes[i].n << "\n";
            output << "o"<< i+1 <<"=" << axes[i].o << "\n";
            output << "d"<< i+1 <<"=" << axes[i].d << "\n";
        }
    }
    output.close();
}

std::shared_ptr<hypercube> VECEXT::readHyper(std::string name){
    std::vector<axis> axes;
    int ndim=0;
    std::string line;
    std::ifstream input;
	input.open(name+".header");
    if (input.is_open()){
        getline(input,line);
		line.erase(0,10);
		line.erase(1,4);
        ndim = std::stoi(line.c_str());
    }

    for (int i=0; i<ndim; i++){
        axis ax;
        getline(input,line);
        line.erase(0,3);
        ax.n = std::stoi(line.c_str());
        getline(input,line);
        line.erase(0,3);
        ax.o = std::stof(line.c_str());
        getline(input,line);
        line.erase(0,3);
        ax.d = std::stof(line.c_str());
        axes.push_back(ax);
    }
    input.close();
    std::shared_ptr<hypercube> hyper( new hypercube(axes));
    return hyper;
}

void VECEXT::read(const std::shared_ptr<floatHyper> vec, std::string name){
    unsigned long long n123 = vec->getHyper()->getN123();
    std::ifstream input;
	input.open(name+".bin", std::ios::in | std::ios::binary);
    if (input.is_open())
	    input.read((char*) vec->getVals(), n123*sizeof(float));
	input.close();
}

void VECEXT::copyHyper(std::shared_ptr<float2DReg> vec, const std::shared_ptr<hypercube> hyper){
    std::vector<axis> axes = hyper->getAxes();
    axes[1].n = vec->getHyper()->getAxis(2).n;
    axes[0].n = vec->getHyper()->getAxis(1).n;

    std::shared_ptr<hypercube> h (new hypercube(axes[0], axes[1]));
    vec->setHyper(h);
}

void VECEXT::add(std::shared_ptr<float2DReg> vec, const float alpha){
    unsigned int nz = VECEXT::getnz(vec);
    unsigned int nx = VECEXT::getnx(vec);
    std::shared_ptr<float2D> vals = vec->_mat;    
    for (int ix=0; ix<nx; ix++){
        for (int iz=0; iz<nz ; iz++){
            (*vals)[ix][iz] += alpha;
        }
    }    
}

void VECEXT::inverse(std::shared_ptr<float2DReg> vec, const float eps){
    unsigned int nz = VECEXT::getnz(vec);
    unsigned int nx = VECEXT::getnx(vec);
    std::shared_ptr<float2D> vals = vec->_mat;
    for (int ix=0; ix<nx; ix++){
        for (int iz=0; iz<nz ; iz++){
            (*vals)[ix][iz] = 1 / ((*vals)[ix][iz] + eps);
        }
    }
}

void VECEXT::revert(std::shared_ptr<float2DReg> vec, const bool xRevert, const bool zRevert){
    unsigned int nx = vec->getHyper()->getAxis(2).n;
    unsigned int nz = vec->getHyper()->getAxis(1).n;
    std::shared_ptr<float2DReg> temp = vec->clone();

    if ((zRevert == true) && (xRevert == false)) {
        for (int ix=0; ix<nx; ix++){
            for (int iz=0; iz<nz; iz++){
                (*vec->_mat)[ix][nz-iz-1] = (*temp->_mat)[ix][iz];
            }
        }
    }

   else if ((zRevert == false) && (xRevert == true)) {
        for (int ix=0; ix<nx; ix++){
            for (int iz=0; iz<nz; iz++){
                (*vec->_mat)[nx-ix-1][iz] = (*temp->_mat)[ix][iz];
            }
        }
    }

    else if ((zRevert == true) && (xRevert == true)) {
        for (int ix=0; ix<nx; ix++){
            for (int iz=0; iz<nz; iz++){
                (*vec->_mat)[nx-ix-1][nz-iz-1] = (*temp->_mat)[ix][iz];
            }
        }
    }

    else{
        std::clog << "Nothing to revert. The input is unchanged.\n";
    }
}

void VECEXT::random(std::shared_ptr<floatHyper> vec, float min, float max){
    unsigned long long n123 = vec->getHyper()->getN123();
    float * vals = vec->getVals();
    for (unsigned long long i=0; i<n123; i++){
            vals[i] += min + (max-min)*(rand() % 10000)/10000.0;
    }    
}

double VECEXT::norm2(const std::shared_ptr<floatHyper> vec){
    unsigned long long n123 = vec->getHyper()->getN123();
    float * vals = vec->getVals();

    double sum = 0;
    for (unsigned long long i=0; i<n123; i++){
        sum += vals[i]*vals[i];
    } 
    return sum;   
}

double VECEXT::rms(const std::shared_ptr<float2DReg> vec){
    return sqrt(norm2(vec)/(VECEXT::getnx(vec)*VECEXT::getnz(vec)));
}

std::shared_ptr<float2DReg> VECEXT::transpose(const std::shared_ptr<float2DReg> vec){
    axis X = vec->getHyper()->getAxis(2);
    axis Z = vec->getHyper()->getAxis(1);
    std::shared_ptr<float2DReg> newvec (new float2DReg(X, Z));

    unsigned int nz = Z.n;
    unsigned int nx = X.n;
    std::shared_ptr<float2D> vals = vec->_mat;
    std::shared_ptr<float2D> newvals = newvec->_mat;

    for (int ix=0; ix < nx; ix++){
        for (int iz=0; iz < nz; iz++){
            (*newvals)[iz][ix] = (*vals)[ix][iz];
        }
    }
    return newvec;
}

std::shared_ptr<float2DReg> VECEXT::expand(const std::shared_ptr<float2DReg> vec, bool xExpand, bool zExpand) {
    axis X = vec->getHyper()->getAxis(2);
    axis Z = vec->getHyper()->getAxis(1);
    unsigned int nz = Z.n;
    unsigned int nx = X.n;
    std::shared_ptr<float2D> vals = vec->_mat;

    if ((zExpand == true) && (xExpand == false)) {
        Z.n = 2*nz - 1;
        std::shared_ptr<float2DReg> newvec (new float2DReg(Z, X));
        std::shared_ptr<float2D> newvals = newvec->_mat;

        for (int ix=0; ix < nx; ix++){
            for (int iz=0; iz < nz; iz++){
                (*newvals)[ix][2*iz] = (*vals)[ix][iz];
            }
        }
        return newvec;
    }

    else if ((zExpand == false) && (xExpand == true)) {
        X.n = 2*nx - 1;
        std::shared_ptr<float2DReg> newvec (new float2DReg(Z, X));
        std::shared_ptr<float2D> newvals = newvec->_mat;

        for (int ix=0; ix < nx; ix++){
            for (int iz=0; iz < nz; iz++){
                (*newvals)[2*ix][iz] = (*vals)[ix][iz];
            }
        }
        return newvec;
    }

    else if ((zExpand == true) && (xExpand == true)) {
        X.n = 2*nx - 1;
        Z.n = 2*nz - 1;
        std::shared_ptr<float2DReg> newvec (new float2DReg(Z, X));
        std::shared_ptr<float2D> newvals = newvec->_mat;

        for (int ix=0; ix < nx; ix++){
            for (int iz=0; iz < nz; iz++){
                (*newvals)[2*ix][2*iz] = (*vals)[ix][iz];
            }
        }
        return newvec;
    }

    else{
        std::clog << "Nothing to expand. The input is unchanged.\n";
    }
}

std::shared_ptr<float2DReg> VECEXT::resize(const std::shared_ptr<float2DReg> vec, unsigned int nx, unsigned int nz){
    axis X = vec->getHyper()->getAxis(2);
    axis Z = vec->getHyper()->getAxis(1);
    X.n = nx;
    Z.n = nz;
    std::shared_ptr<float2DReg> newvec (new float2DReg(Z, X));

    std::shared_ptr<float2D> vals = vec->_mat;
    std::shared_ptr<float2D> newvals = newvec->_mat;

    for (int ix=0; ix < std::min(nx,VECEXT::getnx(vec)); ix++){
        for (int iz=0; iz < std::min(nz,VECEXT::getnz(vec)); iz++){
            (*newvals)[ix][iz] = (*vals)[ix][iz];
        }
    }
    return newvec;
}

std::shared_ptr<float2DReg> VECEXT::pad(const std::shared_ptr<float2DReg> vec, unsigned int nx_start, unsigned int nx_end,
                                unsigned int nz_start, unsigned int nz_end){
    
    axis X = vec->getHyper()->getAxis(2);
    axis Z = vec->getHyper()->getAxis(1);
    unsigned int nx = X.n;
    unsigned int nz = Z.n;
    X.n += nx_start + nx_end;
    Z.n += nz_start + nz_end;
    X.o = X.o - nx_start*X.d;
    Z.o = Z.o - nz_start*Z.d;

    std::shared_ptr<float2DReg> newvec (new float2DReg(Z, X));
    newvec->zero();

    std::shared_ptr<float2D> vals = vec->_mat;
    std::shared_ptr<float2D> newvals = newvec->_mat;

    for (int ix=nx_start; ix<nx_start+nx; ix++){
        for (int iz=nz_start; iz<nz_start+nz; iz++){
            (*newvals)[ix][iz] = (*vals)[ix-nx_start][iz-nz_start];
        }
    }
    return newvec;
}

std::shared_ptr<float2DReg> VECEXT::cut(const std::shared_ptr<float2DReg> vec, unsigned int nx_start, unsigned int nx_end,
                                unsigned int nz_start, unsigned int nz_end){
    
    axis X = vec->getHyper()->getAxis(2);
    axis Z = vec->getHyper()->getAxis(1);
    unsigned int nx = X.n;
    unsigned int nz = Z.n;
    X.n = X.n - nx_start - nx_end;
    Z.n = Z.n - nz_start - nz_end;
    X.o = X.o + nx_start*X.d;
    Z.o = Z.o + nz_start*Z.d;

    std::shared_ptr<float2DReg> newvec (new float2DReg(Z, X));
    newvec->zero();

    std::shared_ptr<float2D> vals = vec->_mat;
    std::shared_ptr<float2D> newvals = newvec->_mat;

    for (int ix=0; ix<X.n; ix++){
        for (int iz=0; iz<Z.n; iz++){
            (*newvals)[ix][iz] = (*vals)[ix+nx_start][iz+nz_start];
        }
    }
    return newvec;
}



std::shared_ptr<float1DReg> VECEXT::stackx(const std::shared_ptr<float2DReg> vec){
    unsigned int nz = VECEXT::getnz(vec);
    unsigned int nx = VECEXT::getnx(vec);
    axis Z = vec->getHyper()->getAxis(1);
    std::shared_ptr<float1DReg> newvec (new float1DReg(Z));

    std::shared_ptr<float2D> vals = vec->_mat;
    std::shared_ptr<float1D> newvals = newvec->_mat;

    for (int ix=0; ix < nx; ix++){
        for (int iz=0; iz < nz; iz++){
            (*newvals)[iz] += (*vals)[ix][iz];
        }
    }
    newvec->scale(1.0/nx);

    return newvec;
}

std::shared_ptr<float1DReg> VECEXT::stackz(const std::shared_ptr<float2DReg> vec){
    unsigned int nz = VECEXT::getnz(vec);
    unsigned int nx = VECEXT::getnx(vec);
    axis Z = vec->getHyper()->getAxis(1);
    std::shared_ptr<float1DReg> newvec (new float1DReg(Z));

    std::shared_ptr<float2D> vals = vec->_mat;
    std::shared_ptr<float1D> newvals = newvec->_mat;

    for (int ix=0; ix < nx; ix++){
        for (int iz=0; iz < nz; iz++){
            (*newvals)[ix] += (*vals)[ix][iz];
        }
    }
    newvec->scale(1.0/nz);

    return newvec;
}

void VECEXT::lpSinc(std::shared_ptr<float2DReg> vec, float wc, int N, float a){

    unsigned int nz = VECEXT::getnz(vec);
    unsigned int nx = VECEXT::getnx(vec);
    if (2*N+1>nz)
        throw std::logic_error("filter longer than data\n");

    std::shared_ptr<float2D> vals = vec->_mat;

    for (int ix=0; ix < nx; ix++){

        // store filter coefficients and needed portion of the data in temporary vectors
        std::vector <float> temp, filt;
        float num;
        
        // apply filtering in the roll-on part
        for (int i=0; i < N; i++){
            temp.push_back((*vals)[ix][i]);
            (*vals)[ix][i]=0;
            num = wc*sinc(wc*M_PI*(i-N))*(a +(1-a)*cos(M_PI*(i-N)/N));
            filt.push_back(num);
        }

        for (int i=N; i < 2*N+1; i++){			
            temp.push_back((*vals)[ix][i]);
            (*vals)[ix][i]=0;
            num = wc*sinc(wc*M_PI*(i-N))*(a +(1-a)*cos(M_PI*(i-N)/N));
            filt.push_back(num);
            for (int j=0; j<=i; j++){
                (*vals)[ix][i-N] = (*vals)[ix][i-N] + temp[i-j]*filt[j];
            }
        }

        // apply filtering to the rest of the data
        for (int i=2*N+1; i<nz; i++){
            temp.erase(temp.begin());
            temp.push_back((*vals)[ix][i]);
            (*vals)[ix][i-N]=0;
            for (int j=0; j<2*N+1; j++){
                (*vals)[ix][i-N] = (*vals)[ix][i-N] + temp[2*N-j]*filt[j];
            }
        }

        // apply filtering in the roll-off part
        for (int i=nz; i < N+nz; i++){
            temp.erase(temp.begin());
            filt.erase(filt.begin());
            (*vals)[ix][i-N]=0;
            for (int j=0; j<temp.size(); j++){
                (*vals)[ix][i-N] = (*vals)[ix][i-N] + temp[temp.size()-j]*filt[j];
            } 
        }
    }
}

void VECEXT::iirButterworth(std::shared_ptr<float2DReg> vec, float wc, bool lp, unsigned int ho){

    if (lp == false) wc = 1-wc;

    // Initialization
    unsigned int nz = VECEXT::getnz(vec);
    unsigned int nx = VECEXT::getnx(vec);
    std::shared_ptr<float2D> vals = vec->_mat;

    int M = 2*ho;
    float Omega_c = tan(M_PI*wc/2);
    float a0, a1, a2, b1, b2, c;
    float xi, xi1, xi2;

    // Apply the same filter ho times
    for (int k=0; k<ho; k++){

        // compute filter coefficients for Low-Pass
        c=1+2*cos(M_PI*(2*k+1)/(2*M))*Omega_c + Omega_c*Omega_c;
        a0=Omega_c*Omega_c/c;
        a1 = 2 * a0;
        a2 = a0;
        b1=2*(Omega_c*Omega_c-1)/c;
        b2=(1-2*cos(M_PI*(2*k+1)/(2*M))*Omega_c +Omega_c*Omega_c)/c;

        // deduce filter coefficients for High-Pass
        if (lp == false){
            a0 = a0;
            a1 = - a1;
            a2 = a2;
            b1 = - b1;
            b2 = b2;
        }

        // apply filter recursively
        for (int ix=0; ix < nx; ix++){
            xi = (*vals)[ix][1];
            xi1 = (*vals)[ix][0];
            xi2 = 0;
            (*vals)[ix][0] = a0*xi1;
            (*vals)[ix][1] = a0*xi + a1*xi1 - b1*(*vals)[ix][0];

            for (int iz=2; iz<nz; iz++){
                xi2 = xi1;
                xi1 = xi;			
                xi = (*vals)[ix][iz];
                (*vals)[ix][iz] = a0*xi + a1*xi1 + a2*xi2 
                        - b1*(*vals)[ix][iz-1] -b2*(*vals)[ix][iz-2];
            }
        }
    }
}

void VECEXT::iirBpButterworth(std::shared_ptr<float2DReg> vec, float wcl, float wch, unsigned int ho){
    
    // Initialization
    unsigned int nz = VECEXT::getnz(vec);
    unsigned int nx = VECEXT::getnx(vec);
    std::shared_ptr<float2D> vals = vec->_mat;

    int M = 2*ho;
    float Omega_ch = tan(M_PI*wch/2);
    float Omega_cl = tan(M_PI*(1-wcl)/2);
    float a0h, a1h, a2h, b1h, b2h, ch;
    float a0l, a1l, a2l, b1l, b2l, cl;
    float a0, a1, a2, a3, a4, b1, b2, b3, b4;
    float xi, xi1, xi2, xi3, xi4;

    // Apply the same filter ho times
    for (int k=0; k<ho; k++){

        // compute filter coefficients for Low-Pass
        ch=1+2*cos(M_PI*(2*k+1)/(2*M))*Omega_ch + Omega_ch*Omega_ch;
        a0h=Omega_ch*Omega_ch/ch;
        a1h = 2 * a0h;
        a2h = a0h;
        b1h=2*(Omega_ch*Omega_ch-1)/ch;
        b2h=(1-2*cos(M_PI*(2*k+1)/(2*M))*Omega_ch +Omega_ch*Omega_ch)/ch;

        // compute filter coefficients for High-Pass
        cl=1+2*cos(M_PI*(2*k+1)/(2*M))*Omega_cl + Omega_cl*Omega_cl;
        a0l=Omega_cl*Omega_cl/cl;
        a1l = - 2 * a0l;
        a2l = a0l;
        b1l=-2*(Omega_cl*Omega_cl-1)/cl;
        b2l=(1-2*cos(M_PI*(2*k+1)/(2*M))*Omega_cl +Omega_cl*Omega_cl)/cl;

        // deduce filter coefficients for Band-Pass
        a0 = a0h*a0l;
        a1 = a0h*a1l + a0l*a1h;
        a2 = a0h*a2l + a1h*a1l + a2h*a0l;
        a3 = a1h*a2l + a2h*a1l;
        a4 = a2h*a2l;
        b1 = b1l + b1h;
        b2 = b2l + b1h*b1l + b2h;
        b3 = b1h*b2l + b2h*b1l;
        b4 = b2h*b2l;
        
        // apply filter recursively
        for (int ix=0; ix < nx; ix++){
            xi = (*vals)[ix][3];
            xi1 = (*vals)[ix][2];
            xi2 = (*vals)[ix][1];
            xi3 = (*vals)[ix][0];
            xi4 = 0;
            (*vals)[ix][0] = a0*xi3;
            (*vals)[ix][1] = a0*xi2 + a1*xi3 - b1*(*vals)[ix][0];
            (*vals)[ix][2] = a0*xi1 + a1*xi2 + a2*xi3 - b1*(*vals)[ix][1] - b2*(*vals)[ix][0];
            (*vals)[ix][3] = a0*xi + a1*xi1 + a2*xi2 + a3*xi3 - b1*(*vals)[ix][2] - b2*(*vals)[ix][1] - b3*(*vals)[ix][0];

            for (int iz=4; iz<nz; iz++){
                xi4 = xi3;
                xi3 = xi2;
                xi2 = xi1;
                xi1 = xi;			
                xi = (*vals)[ix][iz];
                (*vals)[ix][iz] = a0*xi + a1*xi1 + a2*xi2 + a3*xi3 + a4*xi4
                        - b1*(*vals)[ix][iz-1] -b2*(*vals)[ix][iz-2] -b3*(*vals)[ix][iz-3] -b4*(*vals)[ix][iz-4];
            }
        }
    }
}

void VECEXT::fkFilter(std::shared_ptr<float2DReg> vec, float kLow, float kHigh, float taper){
    if ((abs(kLow)>0.5) || (abs(kHigh)>0.5) || (kLow>kHigh))
        throw std::logic_error("The cutoff wavenumbers must obey -0.5 < kLow < kHigh < 0.5");

    fkTransform fk(vec->getHyper());
    axis X = vec->getHyper()->getAxis(2);
    axis Z = vec->getHyper()->getAxis(1);
    Z.n = Z.n /2 +1;
    X.d = fk.dk();
    X.o = fk.ok();
    Z.d = fk.df();
    Z.o = 0;
    std::shared_ptr<complex2DReg> cvec (new complex2DReg(Z, X));
    fk.forward(false,vec,cvec);

    int nx1 = (0.5+kLow-taper)*X.n;
    nx1 = std::max(nx1, 0);
    int nx2 = (0.5+kLow)*X.n;
    int nx3 = std::ceil((0.5+kHigh)*X.n);
    int nx4 = (0.5+kHigh+taper)*X.n;
    nx4 = std::min(nx4, X.n);
    
    for (int ix=0; ix<nx1; ix++){
        for (int iz=0; iz<Z.n; iz++){
            (*cvec->_mat)[ix][iz] = 0;
        }
    }
    
    for (int ix=nx1; ix<nx2; ix++){
        for (int iz=0; iz<Z.n; iz++){
            (*cvec->_mat)[ix][iz] *= std::cos(0.5*M_PI*(kLow + 0.5 - (1.0*ix)/X.n)/taper);
        }
    }
    
    for (int ix=nx3; ix<nx4; ix++){
        for (int iz=0; iz<Z.n; iz++){
            (*cvec->_mat)[ix][iz] *= std::cos(0.5*M_PI*((1.0*ix)/X.n - 0.5 - kHigh)/taper);
        }
    }
    
    for (int ix=nx4; ix<X.n; ix++){
        for (int iz=0; iz<Z.n; iz++){
            (*cvec->_mat)[ix][iz] = 0;
        }
    }
    
    fk.inverse(false,vec,cvec);
}
