#include "pef2D.h"
#include "selection2D.h"
#include "convolution2D.h"
#include "chainOper2D.h"


void pef2D::estimate(std::shared_ptr<float2DReg> pef, const std::shared_ptr<float2DReg> dat, cgls &cg, const bool crossBounds){

    if (_lead > pef->getHyper()->getAxis(1).n-1)
            throw std::logic_error("The lead index exceeds nz");

    if (_lead + _gap > pef->getHyper()->getAxis(1).n-1) 
        throw std::logic_error("Gap too large or filter too short");

    if ((VECEXT::getnx(pef) > VECEXT::getnx(dat)) || (VECEXT::getnz(pef) > VECEXT::getnz(dat)))
        throw std::logic_error("data is smaller than filter");

    // copy the hypercube from data to the PEF without the size and set originx to zero and originz according to the leading 1
    axis X = dat->getHyper()->getAxis(2);
    axis Z = dat->getHyper()->getAxis(1);
    X.o = 0;
    Z.o = -Z.d * _lead;
    std::shared_ptr<hypercube> hyper (new hypercube(Z, X));
    VECEXT::copyHyper(pef, hyper);

    // create extended data vector
    X = dat->getHyper()->getAxis(2);
    Z = dat->getHyper()->getAxis(1);
    X.n += pef->getHyper()->getAxis(2).n -1;
    Z.n += pef->getHyper()->getAxis(1).n -1;
    Z.o = Z.o -Z.d * _lead;
    std::shared_ptr<float2DReg> dataExt (new float2DReg(Z, X));

    // build an equations selector to discard boundaries in the PEF estimation if required
    {
    std::shared_ptr<float2DReg> dat2 (new float2DReg(dat->getHyper()));
    pef->zero();
    dat2->zero();
    VECEXT::add(pef, 1.0);
    VECEXT::add(dat2, 1.0);
    convolution2D conv(pef, dat2->getHyper(), dataExt->getHyper());
    conv.forward(false, dat2, dataExt);
    }

    float N = pef->getHyper()->getAxis(2).n * pef->getHyper()->getAxis(1).n;
    for (int ix=0; ix<X.n; ix++){
        for (int iz=0; iz<Z.n; iz++){
            if ((*dataExt->_mat)[ix][iz] != N) (*dataExt->_mat)[ix][iz] = 0.0;
            else (*dataExt->_mat)[ix][iz] = 1.0;
        }
    }

    selection2D selectEquation(dataExt);
    dataExt->zero();

    // set the pef to zero except for the leading 1
    pef->zero();
    (*pef->_mat)[0][_lead] = 1;

    // define the data convolution matrix
    convolution2D conv(dat, pef->getHyper(), dataExt->getHyper());

    // build the selection operators to separate PEF known and unknown coefficients
    std::shared_ptr<float2DReg> selector (new float2DReg(pef->getHyper()));
    for (int i=0; i<= _lead + _gap; i++)
        (*selector->_mat)[0][i] = 1;

    selection2D selectKnown(selector);
    selection2D selectUnknown(selector);
    selectUnknown.reverse();

    std::shared_ptr<float2DReg> KnP (new float2DReg(pef->getHyper()));
    selectKnown.forward(false, pef, KnP);
    conv.forward(false, KnP, dataExt);
    dataExt->scale(-1);

    // initialize the CGLS solver
    unsigned int niter = VECEXT::getnx(pef)*VECEXT::getnz(pef) - _lead - _gap -1;
    cg.setNiter(niter);

    std::clog << "Maximum number of iterations for CG solver set to: " <<niter<< "\n";

    chainOper2D chain(&conv, &selectUnknown);

    // Choose whether or not to discard the equations at the bounds and run the solver to estimate the PEF
    if (crossBounds == true){
        cg.run(&chain, pef, dataExt);
    }
    else{
        chainOper2D chain2(&selectEquation, &chain);
        cg.run(&chain2, pef, dataExt);
    }
}


void nonStatPef2D::apply(std::shared_ptr<float2DReg> pef, const std::shared_ptr<float2DReg> dat, std::shared_ptr<float2DReg> res, const bool crossBounds){

    if (_lead > pef->getHyper()->getAxis(1).n-1)
            throw std::logic_error("The lead index exceeds nz");

    if (_lead + _gap > pef->getHyper()->getAxis(1).n-1) 
        throw std::logic_error("Gap too large or filter too short");

    if ((VECEXT::getnx(pef) > VECEXT::getnx(dat)) || (VECEXT::getnz(pef) > VECEXT::getnz(dat)))
        throw std::logic_error("data is smaller than filter");

    // check if data and residuals are of the same space and initialize residuals to the data
    dat->checkSame(res);

    // set the hypercube for the PEF, set originx to zero and originz according to the leading 1
    axis X = dat->getHyper()->getAxis(2);
    axis Z = dat->getHyper()->getAxis(1);
    X.o = 0;
    Z.o = -Z.d * _lead;
    std::shared_ptr<hypercube> hyper (new hypercube(Z, X));
    VECEXT::copyHyper(pef, hyper);

    // set the pef to zero except for the leading 1
    pef->zero();
    std::shared_ptr<float2DReg> dpef = pef->clone();
    (*pef->_mat)[0][_lead] = 1;

    // define arrays size
    unsigned int nx_p = pef->getHyper()->getAxis(2).n;
    unsigned int nz_p = pef->getHyper()->getAxis(1).n;
    unsigned int nx_d = dat->getHyper()->getAxis(2).n;
    unsigned int nz_d = dat->getHyper()->getAxis(1).n;

    // define intermediate vectors
    std::shared_ptr<float2DReg> newpef;
    std::shared_ptr<float2DReg> pefz;
    std::shared_ptr<float2DReg> pefx;
    std::shared_ptr<float2DReg> dpefz = pef->clone();
    std::shared_ptr<float2DReg> dpefx = pef->clone();

    double res_temp, res_pefz, res_pefx, data_norm, step_length, step_lengthx, step_lengthz, weightx, weightz;
    

    std::clog << "Starting the non-stationary PEF estimation/application\n";

    // Estimation and application including the boundaries
    if (crossBounds == true){

        // store an array of pef of the same size as nz_d
        std::vector<std::shared_ptr<float2DReg> > pef_array;
        pef_array.push_back(pef->clone());

        // loop over the 1st column of data
        int max_nz_p = 0;
        for (unsigned int iz_r=1; iz_r<nz_d; iz_r++){

            // compute residuals
            res_temp = 0;
            max_nz_p = std::min(nz_p, iz_r + _lead + 1);
            for (int iz_p=_lead; iz_p<max_nz_p; iz_p++){
                res_temp += (*pef->_mat)[0][iz_p] * (*dat->_mat)[0][iz_r - iz_p + _lead];
            }

            // compute filter perturbation
            dpef->zero();
            data_norm = 0;
            for (int iz_p=_lead + 1; iz_p<max_nz_p; iz_p++){
                (*dpef->_mat)[0][iz_p] += res_temp * (*dat->_mat)[0][iz_r - iz_p + _lead];
                data_norm += (*dat->_mat)[0][iz_r - iz_p + _lead] * (*dat->_mat)[0][iz_r - iz_p + _lead];
            }

            // update the pef
            newpef = pef->clone();
            step_length = res_temp*res_temp/(res_temp*res_temp*data_norm + _epsilon*_epsilon);
            step_length = _switch * step_length + (1 - _switch)*_epsilon*_epsilon;
            newpef->scaleAdd(dpef, 1.0, - step_length);

            // update the residuals
            (*res->_mat)[0][iz_r] = 0;
            for (int iz_p=_lead; iz_p<max_nz_p; iz_p++){
                (*res->_mat)[0][iz_r] += (*newpef->_mat)[0][iz_p] * (*dat->_mat)[0][iz_r - iz_p + _lead];
            }

            // store the new pef
            pef_array.push_back(newpef);
            pef = newpef;
        }

        // loop over the rest of the data
        int max_nx_p = 0;
        int min_nz_p = 0;
        for (unsigned int ix_r=1; ix_r<nx_d; ix_r++){

            // treat the first row separately (iz=0)
            pef = pef_array[0];

            // compute residuals
            res_temp = 0;
            max_nx_p = std::min(nx_p, ix_r + 1);
            for (int ix_p=0; ix_p<max_nx_p; ix_p++){
                for (int iz_p=0; iz_p<_lead+1; iz_p++){
                    res_temp += (*pef->_mat)[ix_p][iz_p] * (*dat->_mat)[ix_r - ix_p][ - iz_p + _lead];
                }
            }

            // compute filter perturbation
            dpef->zero();
            data_norm = 0;
            for (int ix_p=0; ix_p<max_nx_p; ix_p++){
                for (int iz_p=0; iz_p<_lead+1; iz_p++){
                    (*dpef->_mat)[ix_p][iz_p] += res_temp * (*dat->_mat)[ix_r - ix_p][ - iz_p + _lead];
                    data_norm += (*dat->_mat)[ix_r - ix_p][ - iz_p + _lead] * (*dat->_mat)[ix_r - ix_p][ - iz_p + _lead];
                }
            }
            for (int iz_p=0; iz_p<_lead+1; iz_p++){
                (*dpef->_mat)[0][iz_p] = 0;
                data_norm -= (*dat->_mat)[ix_r][ - iz_p + _lead] * (*dat->_mat)[ix_r][ - iz_p + _lead];
            }

            // update the pef
            newpef = pef->clone();
            step_length = res_temp*res_temp/(res_temp*res_temp*data_norm + _epsilon*_epsilon);
            step_length = _switch * step_length + (1 - _switch)*_epsilon*_epsilon;
            newpef->scaleAdd(dpef, 1.0, - step_length);

            // update the residuals
            (*res->_mat)[ix_r][0] = 0;
            for (int ix_p=0; ix_p<max_nx_p; ix_p++){
                for (int iz_p=0; iz_p<_lead+1; iz_p++){
                    (*res->_mat)[ix_r][0] += (*newpef->_mat)[ix_p][iz_p] * (*dat->_mat)[ix_r - ix_p][ - iz_p + _lead];
                }
            }

            // store the new pef
            pef_array.erase(pef_array.begin());
            pef_array.push_back(newpef);

            // treat the rest of the rows (iz>0)
            for (unsigned int iz_r=1; iz_r<nz_d; iz_r++){

                pefz = pef_array[nz_d-1];
                pefx = pef_array[0];

                // compute residuals
                res_pefz = 0;
                res_pefx = 0;
                max_nx_p = std::min(nx_p, ix_r + 1);
                max_nz_p = std::min(nz_p, iz_r + _lead + 1);
                min_nz_p = std::max(0.0, 1.0*iz_r + 1 + _lead - nz_d);
                for (int ix_p=0; ix_p<max_nx_p; ix_p++){
                    for (int iz_p=min_nz_p; iz_p<max_nz_p; iz_p++){
                        res_pefz += (*pefz->_mat)[ix_p][iz_p] * (*dat->_mat)[ix_r - ix_p][iz_r - iz_p + _lead];
                        res_pefx += (*pefx->_mat)[ix_p][iz_p] * (*dat->_mat)[ix_r - ix_p][iz_r - iz_p + _lead];
                    }
                }

                // compute filter perturbation
                dpefx->zero();
                dpefz->zero();
                data_norm = 0;
                for (int ix_p=0; ix_p<max_nx_p; ix_p++){
                    for (int iz_p=min_nz_p; iz_p<max_nz_p; iz_p++){
                        (*dpefz->_mat)[ix_p][iz_p] += res_pefz * (*dat->_mat)[ix_r - ix_p][iz_r - iz_p + _lead];
                        (*dpefx->_mat)[ix_p][iz_p] += res_pefx * (*dat->_mat)[ix_r - ix_p][iz_r - iz_p + _lead];
                        data_norm += (*dat->_mat)[ix_r - ix_p][iz_r - iz_p + _lead] * (*dat->_mat)[ix_r - ix_p][iz_r - iz_p + _lead];
                    }
                }

                for (int iz_p=min_nz_p; iz_p<_lead+1; iz_p++){
                    (*dpefx->_mat)[0][iz_p] = 0;
                    (*dpefz->_mat)[0][iz_p] = 0;
                    data_norm -= (*dat->_mat)[ix_r][iz_r - iz_p + _lead] * (*dat->_mat)[ix_r][iz_r - iz_p + _lead];
                }

                // update the pef
                newpef = pef->clone();
                step_lengthx = res_pefx*res_pefx/(res_pefx*res_pefx*data_norm + _epsilon*_epsilon);
                step_lengthz = res_pefz*res_pefz/(res_pefz*res_pefz*data_norm + _epsilon*_epsilon);
                step_lengthx = _switch * step_lengthx + (1 - _switch)*_epsilon*_epsilon;
                step_lengthz = _switch * step_lengthz + (1 - _switch)*_epsilon*_epsilon;
                weightz = (res_pefx*res_pefx + _lambda*_lambda)/(res_pefx*res_pefx + res_pefz*res_pefz + 2*_lambda*_lambda);
                weightz = (1 - _forceWeights)*weightz + _forceWeights * _weightz * _weightz / (_weightz * _weightz + _weightx * _weightx);
                weightx = 1 - weightz;
                for (int ix_p=0; ix_p<nx_p; ix_p++){
                    for (int iz_p=0; iz_p<nz_p; iz_p++){
                        (*newpef->_mat)[ix_p][iz_p] = 0;
                        (*newpef->_mat)[ix_p][iz_p] = weightx * ((*pefx->_mat)[ix_p][iz_p] - step_lengthx*(*dpefx->_mat)[ix_p][iz_p])
                                                    + weightz * ((*pefz->_mat)[ix_p][iz_p] - step_lengthz*(*dpefz->_mat)[ix_p][iz_p]);
                    }
                }

                // update the residuals
                (*res->_mat)[ix_r][iz_r] = 0;
                for (int ix_p=0; ix_p<max_nx_p; ix_p++){
                    for (int iz_p=min_nz_p; iz_p<max_nz_p; iz_p++){
                        (*res->_mat)[ix_r][iz_r] += (*newpef->_mat)[ix_p][iz_p] * (*dat->_mat)[ix_r - ix_p][iz_r - iz_p + _lead];
                    }
                }

                // store the new pef
                pef_array.erase(pef_array.begin());
                pef_array.push_back(newpef);
            }
            //std::clog<<"Residuals norm = " << VECEXT::norm2(res) << "\n";
        }
        pef = pef_array[nz_d - 1]->clone();
    }

    // Estimation and application excluding the boundaries
    else {
    }
    std::clog << "Finishing the non-stationary PEF estimation/application\n";
}