#include <fstream>
#include <chrono>
#include <thread>
#include "cgls.h"
#include "float2DReg.h"
#include "matrixMult.h"
#include "floatHyperExt.h"
#include "gnuplot-iostream.h"

using namespace SEP;

void cgls::run(const oper1D * op, std::shared_ptr<float1DReg> mod, const std::shared_ptr<float1DReg> dat){
    
    std::clog << "Running CG solver\n";

    //data residuals
    std::shared_ptr<float1DReg> dr = dat->clone();
    dr->scale(-1);
    op->forward(true,mod,dr);

    //gradients
    std::shared_ptr<float1DReg> g (new float1DReg(mod->getHyper()));
    std::shared_ptr<float1DReg> g1 (new float1DReg(mod->getHyper()));
    op->adjoint(false,g,dr);

    //conjugate direction
    std::shared_ptr<float1DReg> p (new float1DReg(mod->getHyper()));

    //directions weights
    float alpha = 0;
    float beta = 0;

    //objective function
    _func.clear();
    _func.push_back(VECEXT::norm2(dr));

    //iteration number
    unsigned int k = 0;

    //convergence rate
    float rate = 1;

    // temporary containers
    std::shared_ptr<float1DReg> d (new float1DReg(dat->getHyper()));
    std::shared_ptr<float1DReg> g_temp ;

    unsigned int n = mod->getHyper()->getN123();
    unsigned int nd = dat->getHyper()->getN123();

    //start the CG loop consisting of 7 steps
    while (k<_niter & rate>_threshold){

        //step 1: p = -g + beta*p
        for (int i=0; i<n; i++){
            (*p->_mat)[i] = beta*(*p->_mat)[i] - (*g->_mat)[i];
        }

        //step 2: compute alpha
        op->forward(false,p,d);
        alpha = VECEXT::norm2(g) / VECEXT::norm2(d);

        //step 3: update model
        for (int i=0; i<n; i++){
            (*mod->_mat)[i] += alpha*(*p->_mat)[i];
        }

        //step 4: update data residuals
        for (int i=0; i<nd; i++){
            (*dr->_mat)[i] += alpha*(*d->_mat)[i];
        }
        //store the normalized data residuals
        _func.push_back(VECEXT::norm2(dr)/_func[0]);

        //compute the rate of convergence
        if(k>0){
            rate = (_func[k]-_func[k+1])/_func[k];
            if((_func[k]<1e-07) || (_func[k]-_func[k+1]<1e-07)) rate = 0;
        }

        //step 5: update gradient
        op->adjoint(false,g1,dr);

        //step 6: compute beta
        beta = VECEXT::norm2(g1) / VECEXT::norm2(g);

        //copy new gradient into previous one
        g_temp = g;
        g = g1;
        g1 = g_temp;

        //step 7: iterate
        k++;
    }
    _func[0]=1;

    std::clog << "Total number of CG iterations: "<<k<<"\n";
}

void cgls::run(const oper2D * op, std::shared_ptr<float2DReg> mod, const std::shared_ptr<float2DReg> dat){
    
    std::clog << "Running CG solver\n";

    //data residuals
    std::shared_ptr<float2DReg> dr = dat->clone();
    dr->scale(-1);
    op->forward(true,mod,dr);

    //gradients
    std::shared_ptr<float2DReg> g (new float2DReg(mod->getHyper()));
    std::shared_ptr<float2DReg> g1 (new float2DReg(mod->getHyper()));
    op->adjoint(false,g,dr);

    //conjugate direction
    std::shared_ptr<float2DReg> p (new float2DReg(mod->getHyper()));

    //directions weights
    float alpha = 0;
    float beta = 0;

    //objective function
    _func.clear();
    _func.push_back(VECEXT::norm2(dr));

    //iteration number
    unsigned int k = 0;

    //convergence rate
    float rate = 1;

    // temporary containers
    std::shared_ptr<float2DReg> d (new float2DReg(dat->getHyper()));
    std::shared_ptr<float2DReg> g_temp ;

    unsigned int n = mod->getHyper()->getN123();
    unsigned int nd = dat->getHyper()->getN123();

    //start the CG loop consisting of 7 steps
    while (k<_niter & rate>_threshold){

        //step 1: p = -g + beta*p
        for (int i=0; i<n; i++){
            (p->getVals())[i] = beta*(p->getVals())[i] - (g->getVals())[i];
        }

        //step 2: compute alpha
        op->forward(false,p,d);
        alpha = VECEXT::norm2(g) / VECEXT::norm2(d);

        //step 3: update model
        for (int i=0; i<n; i++){
            (mod->getVals())[i] += alpha*(p->getVals())[i];
        }

        //step 4: update data residuals
        for (int i=0; i<nd; i++){
                (dr->getVals())[i] += alpha*(d->getVals())[i];
        }

        //store the normalized data residuals
        _func.push_back(VECEXT::norm2(dr)/_func[0]);

        //compute the rate of convergence
        if(k>0){
            rate = (_func[k]-_func[k+1])/_func[k];
            if((_func[k]<1e-07) || (_func[k]-_func[k+1]<1e-07)) rate = 0;
        }

        //step 5: update gradient
        op->adjoint(false,g1,dr);

        //step 6: compute beta
        beta = VECEXT::norm2(g1) / VECEXT::norm2(g);

        //copy new gradient into previous one
        g_temp = g;
        g = g1;
        g1 = g_temp;

        //step 7: iterate
        k++;
    }

    _func[0]=1;

    std::clog << "Total number of CG iterations: "<<k<<"\n";
}

void cgls::test(){

    matrixMult mult(50,200);
    std::shared_ptr<float1DReg> mod (new float1DReg(50));
    std::shared_ptr<float1DReg> mod0 (new float1DReg(50));
    std::shared_ptr<float1DReg> dat (new float1DReg(200));

    mult.random(-1,1);
    VECEXT::random(mod,-10,10);
    mult.forward(false,mod,dat);

    _niter = 50;
    _threshold = 0.0;

    this->run(&mult, mod0, dat);

    // L2 squared error of estimated model
    double trueNorm, estimatedNorm, dataNorm, error;
    trueNorm = VECEXT::norm2(mod);
    estimatedNorm = VECEXT::norm2(mod0);
    dataNorm = VECEXT::norm2(dat);
    mod0->scale(-1);
    mod->add(mod0);
    error = VECEXT::norm2(mod);

    std::clog << "data norm squared = " << dataNorm << std::endl;
    std::clog << "true model norm squared = " << trueNorm << std::endl;
    std::clog << "estimated model norm squared = " << estimatedNorm << std::endl;
    std::clog << "model error squared = " << error << std::endl;
}

void cgls::plotFunc(bool logscale){

    Gnuplot gp;
    gp << "set title \"Normalized objective function\"\n";
    gp << "set xlabel \"Iteration number\" \n";
    if (logscale == true){
        gp << "set logscale y\n";
    }
    gp << "plot '-' with linespoints notitle\n";
    gp.send1d(_func);
}

void cgls::saveFunc(bool logscale, std::string name, std::string directory){

    Gnuplot gp;
    gp << "set terminal png\n";
    gp << "set output '"<< directory << name << ".png'\n";
    gp << "set title \"Normalized objective function\"\n";
    gp << "set xlabel \"Iteration number\" \n";
    if (logscale == true){
        gp << "set logscale y\n";
    }
    gp << "plot '-' with linespoints notitle\n";
    gp.send1d(_func);
}