#include <fstream>
#include <chrono>
#include <thread>
#include "cglsReg.h"
#include "float2DReg.h"
#include "matrixMult.h"
#include "identity1D.h"

using namespace SEP;

void cglsReg::run(const oper1D * L, const oper1D * D, const float eps, std::shared_ptr<float1DReg> mod, const std::shared_ptr<float1DReg> dat, std::shared_ptr<float1DReg> dat2){
    //initialize variables

    //extended data residuals
    std::shared_ptr<float1DReg> drd = dat->clone();
    drd->scale(-1);
    L->forward(true,mod,drd);
    if (dat2 == nullptr)
        dat2 = std::make_shared<float1DReg>(D->getRange());
    std::shared_ptr<float1DReg> drm = dat2->clone();
    drm->scale(-1);
    D->forward(true,mod,drm);
    drm->scale(eps);

    //gradients
    std::shared_ptr<float1DReg> g (new float1DReg(mod->getHyper()));
    std::shared_ptr<float1DReg> g1 (new float1DReg(mod->getHyper()));
    D->adjoint(false,g,drm);
    g->scale(eps);
    L->adjoint(true,g,drd);
    
    //conjugate direction
    std::shared_ptr<float1DReg> p (new float1DReg(mod->getHyper()));

    //directions weights
    float alpha = 0;
    float beta = 0;

    //objective function
    _func.clear();
    _func.push_back(VECEXT::norm2(drd)+VECEXT::norm2(drm));

    //iteration number
    unsigned int k = 0;

    //convergence rate
    float rate = 1;
    
    // temporary containers
    std::shared_ptr<float1DReg> dd (new float1DReg(dat->getHyper()));
    std::shared_ptr<float1DReg> dm (new float1DReg(D->getRange()));
    std::shared_ptr<float1DReg> g_temp;

    unsigned int n = mod->getHyper()->getN123();
    unsigned int nd = dat->getHyper()->getN123();
    unsigned int nd2 = dat2->getHyper()->getN123();

    //start the CG loop consisting of 7 steps
    while (k<_niter & rate>_threshold){

        //step 1: p = -g + beta*p
        for (int i=0; i<n; i++){
            (p->getVals())[i] = beta*(p->getVals())[i] - (g->getVals())[i];
        }
        
        //step 2: compute alpha
        L->forward(false,p,dd);
        D->forward(false,p,dm);
        dm->scale(eps);
        alpha = VECEXT::norm2(g) / (VECEXT::norm2(dd)+VECEXT::norm2(dm));
        
        //step 3: update model
        for (int i=0; i<n; i++){
            (mod->getVals())[i] += alpha*(p->getVals())[i];
        }
        
        //step 4: update extended data residuals
        for (int i=0; i<nd; i++){
                (drd->getVals())[i] += alpha*(dd->getVals())[i];
        }
        for (int i=0; i<nd2; i++){
            (drm->getVals())[i] += alpha*(dm->getVals())[i];
        }
        //store the normalized extended data residuals
        _func.push_back((VECEXT::norm2(drd)+VECEXT::norm2(drm))/_func[0]);

        //compute the rate of convergence
        if(k>0){
            rate = (_func[k]-_func[k+1])/_func[k];
            if((_func[k]<1e-07) || (_func[k]-_func[k+1]<1e-07)) rate = 0;
        }

        //step 5: update gradient
        D->adjoint(false,g1,drm);
        g1->scale(eps);
        L->adjoint(true,g1,drd);

        //step 6: compute beta
        beta = VECEXT::norm2(g1) / VECEXT::norm2(g);

        //copy new gradient into previous one
        g_temp = g;
        g = g1;
        g1 = g_temp;

        //step 7: iterate
        k++;
    }

    _func[0]=1;
}

void cglsReg::run(const oper2D * L, const oper2D * D, const float eps, std::shared_ptr<float2DReg> mod, const std::shared_ptr<float2DReg> dat, std::shared_ptr<float2DReg> dat2){
    //initialize variables

    //extended data residuals
    std::shared_ptr<float2DReg> drd = dat->clone();
    drd->scale(-1);
    L->forward(true,mod,drd);
    if (dat2 == nullptr)
        dat2 = std::make_shared<float2DReg>(D->getRange());
    std::shared_ptr<float2DReg> drm = dat2->clone();
    drm->scale(-1);
    D->forward(true,mod,drm);
    drm->scale(eps);

    //gradients
    std::shared_ptr<float2DReg> g (new float2DReg(mod->getHyper()));
    std::shared_ptr<float2DReg> g1 (new float2DReg(mod->getHyper()));
    D->adjoint(false,g,drm);
    g->scale(eps);
    L->adjoint(true,g,drd);
    
    //conjugate direction
    std::shared_ptr<float2DReg> p (new float2DReg(mod->getHyper()));

    //directions weights
    float alpha = 0;
    float beta = 0;

    //objective function
    _func.clear();
    _func.push_back(VECEXT::norm2(drd)+VECEXT::norm2(drm));

    //iteration number
    unsigned int k = 0;

    //convergence rate
    float rate = 1;

    // temporary containers
    std::shared_ptr<float2DReg> dd (new float2DReg(dat->getHyper()));
    std::shared_ptr<float2DReg> dm (new float2DReg(D->getRange()));
    std::shared_ptr<float2DReg> g_temp;

    unsigned int n = mod->getHyper()->getN123();
    unsigned int nd = dat->getHyper()->getN123();
    unsigned int nd2 = dat2->getHyper()->getN123();

    //start the CG loop consisting of 7 steps
    while (k<_niter & rate>_threshold){

        //step 1: p = -g + beta*p
        for (int i=0; i<n; i++){
            (p->getVals())[i] = beta*(p->getVals())[i] - (g->getVals())[i];
        }
        
        //step 2: compute alpha
        L->forward(false,p,dd);
        D->forward(false,p,dm);
        dm->scale(eps);
        alpha = VECEXT::norm2(g) / (VECEXT::norm2(dd)+VECEXT::norm2(dm));
        
        //step 3: update model
        for (int i=0; i<n; i++){
            (mod->getVals())[i] += alpha*(p->getVals())[i];
        }
        
        //step 4: update extended data residuals
        for (int i=0; i<nd; i++){
                (drd->getVals())[i] += alpha*(dd->getVals())[i];
        }
        for (int i=0; i<nd2; i++){
            (drm->getVals())[i] += alpha*(dm->getVals())[i];
        }

        //store the normalized extended data residuals
        _func.push_back((VECEXT::norm2(drd)+VECEXT::norm2(drm))/_func[0]);

        //compute the rate of convergence
        if(k>0){
            rate = (_func[k]-_func[k+1])/_func[k];
            if((_func[k]<1e-07) || (_func[k]-_func[k+1]<1e-07)) rate = 0;
        }

        //step 5: update gradient
        D->adjoint(false,g1,drm);
        g1->scale(eps);
        L->adjoint(true,g1,drd);

        //step 6: compute beta
        beta = VECEXT::norm2(g1) / VECEXT::norm2(g);

        //copy new gradient into previous one
        g_temp = g;
        g = g1;
        g1 = g_temp;

        //step 7: iterate
        k++;
    }

    _func[0]=1;
}

void cglsReg::test(){

    matrixMult mult(50,200);
    std::shared_ptr<float1DReg> mod (new float1DReg(50));
    std::shared_ptr<float1DReg> mod0 (new float1DReg(50));
    std::shared_ptr<float1DReg> mod1 (new float1DReg(50));
    std::shared_ptr<float1DReg> dat (new float1DReg(200));

    mult.random(-1,1);
    VECEXT::random(mod,-10,10);
    mult.forward(false,mod,dat);

    _niter = 50;
    _threshold = 0.0;
    identity1D Id (mod->getHyper());
    float eps0 = 1e-03;
    float eps1 = 10;

    this->run(&mult, &Id, eps0, mod0, dat);
    this->run(&mult, &Id, eps1, mod1, dat);

    // L2 squared error of estimated model
    double trueNorm, estimatedNorm0, estimatedNorm1, dataNorm, error0, error1;
    dataNorm = VECEXT::norm2(dat);
    trueNorm = VECEXT::norm2(mod);
    estimatedNorm0 = VECEXT::norm2(mod0);
    estimatedNorm1 = VECEXT::norm2(mod1);
    mod->scale(-1);
    mod0->add(mod);
    mod1->add(mod);
    error0 = VECEXT::norm2(mod0);
    error1 = VECEXT::norm2(mod1);

    std::clog << "data norm squared = " << dataNorm << std::endl;
    std::clog << "true model norm squared = " << trueNorm << std::endl;
    std::clog << "estimated model norm squared (weak reg) = " << estimatedNorm0 << std::endl;
    std::clog << "estimated model norm squared (strong reg) = " << estimatedNorm1 << std::endl;
    std::clog << "model error squared (weak reg) = " << error0 << std::endl;
    std::clog << "model error squared (strong reg) = " << error1 << std::endl;
}