#include "Trace.h"
#include "functions.h"
#include <fftw3.h>

	
	// Display the values of the trace
	void Trace::print(string message, ostream &o){
		string espace(message.size(), ' ');
		o << message << " [" ;
		for (int j =0; j < (*vec).size(); j++)
			o << " " << (*vec)[j];
		o << " ]" << endl;
	}

	// Save trace to an ascii file with a special format
	void Trace::save(string file){
		ofstream out; out.open(file);
		out << "#si=" << si << endl;
		out << "#t0=" << t0 << endl;
		TYPE * p = (*vec).data();
		for (int i=0; i<(*vec).size(); i++) {out << p[i] << endl;}
		out.close();
	}
	
	// Compute norm2 squared of the trace
	TYPE Trace::norm2(){
		TYPE norm = 0;
		TYPE * p = (*vec).data();
		for (int i=0; i<(*vec).size(); ++i){norm = norm + p[i]*p[i];}
		return norm;
	}
	
	// Add a constant to the trace
	Trace Trace::operator + (const TYPE c){
		vector <TYPE> * v = new vector<TYPE>();
		unsigned int N = vec->size();
		TYPE * ptr = vec->data();
		for (int i=0; i<N; i++)
			v->push_back(ptr[i]+c);
		Trace tr(v,si,t0);
		return tr;
	}

	// Add two traces together
    Trace Trace::operator + (Trace &tr){
        if ((si != tr.si) || (t0 != tr.t0) || (vec->size() != tr.vec->size()))
            {cout << "Trace mismatch"; return *this;}
        else {
			vector <TYPE> * v = new vector<TYPE>();
            unsigned int Nd = vec->size();
            TYPE * pd = vec->data();
			TYPE * pf = tr.vec->data();
            for (int i=0; i<Nd; i++)
                v->push_back(pd[i]+pf[i]);
            return Trace(v,si,t0);
        }
    }

	// Subtract a constant from the trace
	Trace Trace::operator - (const TYPE c){
		vector <TYPE> * v = new vector<TYPE>();
		unsigned int N = vec->size();
		TYPE * ptr = vec->data();
		for (int i=0; i<N; i++)
			v->push_back(ptr[i]-c);
		return Trace(v,si,t0);
	}

	// Subtract one trace from another
	Trace Trace::operator - (Trace &tr){
        if ((si != tr.si) || (t0 != tr.t0) || (vec->size() != tr.vec->size()))
            {cout << "Trace mismatch"; return *this;}
        else {
			vector <TYPE> * v = new vector<TYPE>();
            unsigned int Nd = vec->size();
            TYPE * pd = vec->data();
			TYPE * pf = tr.vec->data();
            for (int i=0; i<Nd; i++)
                v->push_back(pd[i]-pf[i]);
            return Trace(v,si,t0);
        }
    }

	// Multiply the trace by a constant
	Trace Trace::operator * (const TYPE c){
		vector <TYPE> * v = new vector<TYPE>();
		unsigned int N = vec->size();
		TYPE * ptr = vec->data();
		for (int i=0; i<N; i++)
			v->push_back(ptr[i]*c);
		return Trace(v,si,t0);
	}

	// Divide the trace by a non-zero constant
	Trace Trace::operator / (const TYPE c){
		vector <TYPE> * v = new vector<TYPE>();
		unsigned int N = vec->size();
		TYPE * ptr = vec->data();
		for (int i=0; i<N; i++)
			v->push_back(ptr[i]/c);
		return Trace(v,si,t0);
	}

	// Scale the trace by a constant
	void Trace::scale(const TYPE c){
		unsigned int N = vec->size();
		TYPE * ptr = vec->data();
		for (int i=0; i<N; i++)
			ptr[i]=ptr[i]*c;
	}

	// Scale the trace by another trace
	void Trace::scale(Trace &f){
		if ((si != f.si) || (t0 != f.t0) || (vec->size() != f.vec->size()))
            {cout << "Trace mismatch";}
        else {
			unsigned int N = vec->size();
			TYPE * ptr = vec->data();
			TYPE * pf = f.vec->data();
			for (int i=0; i<N; i++)
				ptr[i]=ptr[i]*pf[i];
		}
	}

    // Dot product between two traces
    TYPE Trace::operator * (Trace &f){
        if ((si != f.si) || (t0 != f.t0) || (vec->size() != f.vec->size()))
            {cout << "Trace mismatch"; return 0;}
        else {
            unsigned int Nd = vec->size();
            TYPE * pd = vec->data();
			TYPE * pf = f.vec->data();
            TYPE sum = 0;
            for (int i=0; i<Nd; i++)
                sum = sum + pd[i]*pf[i];
            return sum;
        }
    }
	
	// Differentiate the trace
	void Trace::differentiate(bool freq_domain){
		if (freq_domain == true){
			vector <complex<TYPE> >* fft = FFT(vec);
			unsigned int N = fft->size();
			TYPE omega = M_PI/(N*si);
			complex<TYPE>* ptr = fft->data();
			// multiply by j.omega.i ; j is the imaginary number
			for (int i=0; i<N; i++){
				complex<TYPE> c(0, omega*i);
				ptr[i] *= c;
			}
			delete vec;
			vec = inv_FFT(fft);
		}
		// Differentiate in time domain
		else {
			TYPE * ptr = vec->data();
			unsigned int N = vec->size();
			for (int i=0; i<N-1; i++){
				ptr[i] = (ptr[i+1] - ptr[i]) / si;
			}
			ptr[N-1] = -ptr[N-1] / si;
		}
	}

	// Integrate the trace
	void Trace::integrate(bool freq_domain){
		if (freq_domain == true){
			vector <complex<TYPE> >* fft = FFT(vec);
			unsigned int N = fft->size();
			TYPE omega = M_PI/(N*si);
			complex<TYPE>* ptr = fft->data();
			// divide by j.omega.i ; j is the imaginary number
			for (int i=1; i<N; i++){
				complex<TYPE> c(0, -1/(omega*i));
				ptr[i] *= c;
			}
			complex<TYPE> c(N*si, 0);
			ptr[0] *= c;
			delete vec;
			vec = inv_FFT(fft);
		}
		// Integrate in time domain
		else {
			TYPE * ptr = vec->data();
			unsigned int N = vec->size();
			ptr[0] *= si;
			for (int i=1; i<N; i++){
				ptr[i] = ptr[i-1] + ptr[i] * si;
			}
		}
	}

	// Interpolate data in time or frequency domain
	void Trace::interp(unsigned int factor, bool freq_domain){
		si = si/(factor+1);
		if (freq_domain == true){
			vector <complex<TYPE> >* fft = FFT(vec);
			unsigned int N = fft->size()-1;
			vector <complex<TYPE> > zero(N*factor);
			if ((*this).size()-N*2 == 1){
				complex <TYPE> c(0,0);
				for (int i=0; i<floor(factor/2)+1; i++){
					zero.push_back(c);
				}
			}
			fft->insert(fft->end(),zero.begin(),zero.end());
			delete vec;
			vec = inv_FFT(fft);
			(*this).scale(factor+1);
		}
		else {
			unsigned int N = vec->size();
			vector <TYPE>* v = new vector<TYPE>(N*(factor+1));
			TYPE * p = vec->data();
			TYPE * pd = v->data();
			for (int i=0; i<N; i++){
				for (int j=0; j<N; j++){
					pd[j*(factor+1)] = p[j];
					for (int k=1; k<=factor; k++){
						pd[j*(factor+1)+k] = pd[j*(factor+1)+k] + p[i]*sinc(M_PI*((j*(factor+1)+k)/(factor+1.0)-i));
					}
				}
			}
			delete vec;
			vec = v;
		}
	}

	// Apply FIR Hanning filter
	void Trace::fir_hanning(){
		TYPE * ptr = vec->data();
		TYPE xk2 = ptr[0], xk1 = ptr[1], xk = ptr[2];
		ptr[1]=1./2 * (xk1 + xk2);
		ptr[2]=1./4 * (xk + 2*xk1 + xk2);
		for (int i=3; i<vec->size(); i++) {
			xk2 = xk1;
			xk1 = xk;
			xk = ptr[i];
			ptr[i]=1./4 * (xk + 2*xk1 + xk2);
		}
	}

	// Apply FIR windowed sinc filter
	// wc is the percentage of Nyquist frequency
	// N is the half length of sinc operator, in number of samples
	// a is the coefficient used for cosine windowing
	void Trace::fir_sinc(TYPE wc, int N, TYPE a){
		unsigned int Nd = vec->size();
		TYPE * ptr = vec->data();
		vector <TYPE> temp, filt;
		TYPE num;
		// 
		for (int i=0; i<2*N+1; i++){			
			temp.push_back(ptr[i]);
			num = wc*sinc(wc*M_PI*(i-N))*(a +(1-a)*cos(M_PI*(i-N)/N));
			filt.push_back(num);
			ptr[i]=0;
			for (int j=0; j<=i; j++){
				ptr[i] = ptr[i] + temp[j]*filt[i-j];
			}
		}
		for (int i=2*N+1; i<Nd; i++){
			temp.erase(temp.begin());
			temp.push_back(ptr[i]);
			ptr[i]=0;
			for (int j=0; j<2*N+1; j++){
				ptr[i] = ptr[i] + temp[j]*filt[2*N-j];
			}
		}
		vec->erase(vec->begin(), vec->begin()+N);
		vec->resize(Nd);
	}

	// Apply IIR Butterworth filter
	// wc is the percentage of Nyquist frequency
	// ho is the half order of the filter
	// visit http://www.ee.ic.ac.uk/hp/staff/dmb/courses/DSPDF/00800_TransIIR.pdf
	// for filters transformation (slide 66)
	void Trace::iir_butterworth(TYPE wc, bool lp, unsigned int ho){
		
		if (lp == false) wc = 1-wc;

		// Initialization
		int M = 2*ho;
		TYPE Omega_c = tan(M_PI*wc/2);
		TYPE a0, a1, a2, b1, b2, c;
		TYPE * ptr = vec->data();
		unsigned int N = vec->size();
		TYPE xi, xi1, xi2;
		// Apply the same filter ho times
		for (int k=0; k<ho; k++){
			// compute filter coefficients for Low-Pass
			c=1+2*cos(M_PI*(2*k+1)/(2*M))*Omega_c + Omega_c*Omega_c;
			a0=Omega_c*Omega_c/c;
			a1 = 2 * a0;
			a2 = a0;
			b1=2*(Omega_c*Omega_c-1)/c;
			b2=(1-2*cos(M_PI*(2*k+1)/(2*M))*Omega_c +Omega_c*Omega_c)/c;

			// deduce filter coefficients for High-Pass
			if (lp == false){
				a0 = a0;
				a1 = - a1;
				a2 = a2;
				b1 = - b1;
				b2 = b2;
			}
			
			// apply filter recursively
			xi = ptr[1];
			xi1 = ptr[0];
			xi2 = 0;
			ptr[0] = a0*xi1;
			ptr[1] = a0*xi + a1*xi1 - b1*ptr[0];
			for (int i=2; i<N; i++){
				xi2 = xi1;
				xi1 = xi;			
				xi = ptr[i];
				ptr[i] = a0*xi + a1*xi1 + a2*xi2 
						- b1*ptr[i-1] -b2*ptr[i-2];
			}
		}
	}

// Apply IIR Butterworth band-pass filter
// wcl is the percentage of Nyquist frequency for the low end
// wch is the percentage of Nyquist frequency for the high end
// ho is the half order of the filter
void Trace::iir_butterworth_bp(TYPE wcl, TYPE wch, unsigned int ho){
	// Initialization
		int M = 2*ho;
		TYPE Omega_ch = tan(M_PI*wch/2);
		TYPE Omega_cl = tan(M_PI*(1-wcl)/2);
		TYPE a0h, a1h, a2h, b1h, b2h, ch;
		TYPE a0l, a1l, a2l, b1l, b2l, cl;
		TYPE a0, a1, a2, a3, a4, b1, b2, b3, b4;
		TYPE * ptr = vec->data();
		unsigned int N = vec->size();
		TYPE xi, xi1, xi2, xi3, xi4;

		// Apply the same filter ho times
		for (int k=0; k<ho; k++){
			// compute filter coefficients for Low-Pass
			ch=1+2*cos(M_PI*(2*k+1)/(2*M))*Omega_ch + Omega_ch*Omega_ch;
			a0h=Omega_ch*Omega_ch/ch;
			a1h = 2 * a0h;
			a2h = a0h;
			b1h=2*(Omega_ch*Omega_ch-1)/ch;
			b2h=(1-2*cos(M_PI*(2*k+1)/(2*M))*Omega_ch +Omega_ch*Omega_ch)/ch;

			// compute filter coefficients for High-Pass
			cl=1+2*cos(M_PI*(2*k+1)/(2*M))*Omega_cl + Omega_cl*Omega_cl;
			a0l=Omega_cl*Omega_cl/cl;
			a1l = - 2 * a0l;
			a2l = a0l;
			b1l=-2*(Omega_cl*Omega_cl-1)/cl;
			b2l=(1-2*cos(M_PI*(2*k+1)/(2*M))*Omega_cl +Omega_cl*Omega_cl)/cl;

			// deduce filter coefficients for Band-Pass
			a0 = a0h*a0l;
			a1 = a0h*a1l + a0l*a1h;
			a2 = a0h*a2l + a1h*a1l + a2h*a0l;
			a3 = a1h*a2l + a2h*a1l;
			a4 = a2h*a2l;
			b1 = b1l + b1h;
			b2 = b2l + b1h*b1l + b2h;
			b3 = b1h*b2l + b2h*b1l;
			b4 = b2h*b2l;
			
			// apply filter recursively
			xi = ptr[3];
			xi1 = ptr[2];
			xi2 = ptr[1];
			xi3 = ptr[0];
			xi4 = 0;
			ptr[0] = a0*xi3;
			ptr[1] = a0*xi2 + a1*xi3 - b1*ptr[0];
			ptr[2] = a0*xi1 + a1*xi2 + a2*xi3 - b1*ptr[1] - b2*ptr[0];
			ptr[3] = a0*xi + a1*xi1 + a2*xi2 + a3*xi3 - b1*ptr[2] - b2*ptr[1] - b3*ptr[0];
			for (int i=4; i<N; i++){
				xi4 = xi3;
				xi3 = xi2;
				xi2 = xi1;
				xi1 = xi;			
				xi = ptr[i];
				ptr[i] = a0*xi + a1*xi1 + a2*xi2 + a3*xi3 + a4*xi4
						- b1*ptr[i-1] -b2*ptr[i-2] -b3*ptr[i-3] -b4*ptr[i-4];
			}
		}
}

// Transform the trace into a zero-phase equivalent
void Trace::zero_phase(){
	int Nd = vec->size();

	// Make the vector length odd
	if (2*(Nd/2)==Nd) {
		Nd+=1;
		vec->push_back(0);
	}

	vector <complex<TYPE> >* fft = FFT(vec);
	complex<TYPE>* ptr = fft->data();
	unsigned int N = fft->size();
	TYPE dw = 2*M_PI/Nd;
	TYPE r;

	// Modify the FFT by setting a linear phase while keeping the same
	// power spectrum
	for (int i=0; i<N; i++){
		r = sqrt(norm(ptr[i]));
		ptr[i].real(r*cos(i*dw*N));
		ptr[i].imag(r*sin(i*dw*N));
	}
	delete vec;
	vec = inv_FFT(fft);
	t0 = - Nd/2;
}


// Convolve a trace with a filter
Trace* convolve(Trace* tr, Trace* f, bool truncate){
	if (tr->get_si() != f->get_si()) {cout << "Sampling interval mismatch\n"; exit(1);}
	else {
		unsigned int Nd = tr->size(), Nf = f->size();
		vector <TYPE> * v = new vector<TYPE>(Nd + Nf - 1);
		Trace * out = new Trace(v);
		out->set_si(tr->get_si()); out->set_t0(tr->get_t0() + f->get_t0());
		TYPE * p = v->data();
		TYPE * pd = tr->get_data();
		TYPE * pf = f->get_data();
		for (int i=0; i<Nd; i++){
			for (int j=0; j<Nf; j++){
				p[i+j] = p[i+j] + pd[i]*pf[j];
			}
		}
        if (truncate == true) out->resize(Nd);
		return out;
	}
}

// Cross correlate a trace with a filter
Trace* correlate(Trace* tr, Trace* f, bool truncate){
	if (tr->get_si() != f->get_si()) {cout << "Sampling interval mismatch\n"; exit(1);}
	else {
		unsigned int Nd = tr->size(), Nf = f->size();
		vector <TYPE> * v = new vector<TYPE>(Nd + Nf - 1);
		Trace * out = new Trace(v);
		out->set_si(tr->get_si()); out->set_t0(tr->get_t0() + f->get_t0());
		TYPE * p = v->data();
		TYPE * pd = tr->get_data();
		TYPE * pf = f->get_data();
		for (int i=0; i<Nd; i++){
			for (int j=0; j<Nf; j++){
				p[i+Nf-j-1] = p[i+Nf-j-1] + pd[i]*pf[j];
			}
		}
		if (truncate == true) out->erase_start(Nf-1);
		return out;
	}
}