#include <math.h>
#include <iostream>
#include "pef_1d.h"
#include "Trace.h"
#include "functions.h"


// Estimate the 1d stationary Predictionary Error Filter PEF from 1D data
// The data has Nd+1 elements
// The PEF has Nf+1 elements (including the leading 1)
// The residuals can be written as r = d - Gm where
// d = (data 0)' Nd+Nf+1 vector
// m = (f1,...,fNf)' the unknown PEF coefficients
// G is the (Nd+Nf+1)xNf convolution matrix depending solely on data


// Compute the action of G on a given vector v of size=Nf
vector <TYPE>* G(vector <TYPE>* data, vector <TYPE>* v){
    unsigned int Nd = data->size()-1;
    unsigned int Nf = v->size();
    vector <TYPE>* out = new vector <TYPE>(Nd+Nf+1);
    TYPE * pd = data->data();
    TYPE * pf = v->data();
    TYPE * p = out->data();
    p[0]=0;
    for (int i=0; i<=Nd; i++){
				for (int j=0; j<Nf; j++){
					p[i+j+1] = p[i+j+1] + pd[i]*pf[j];
				}
			}
    return out;
}

// Compute the action of GT (transpose G) on a given vector v of size Nd+Nf+1
vector <TYPE>* GT(vector <TYPE>* data, vector <TYPE>* v){
    unsigned int Nd = data->size()-1;
    unsigned int Nf = v->size()-Nd-1;
    vector <TYPE>* out = new vector <TYPE>(Nf);
    TYPE * pd = data->data();
    TYPE * pf = v->data();
    TYPE * p = out->data();
    for (int i=0; i<=Nd; i++){
				for (int j=0; j<Nf; j++){
                    p[j] = p[j] + pd[i]*pf[i+j+1];
				}
			}
    return out;
}

// Perform the dot product test for G
void G_dot_product(){
    // initialize data, x, and y vectors with random numbers
    unsigned int Nd=1000;
    unsigned int Nf=100;
    vector <TYPE> * data = new vector<TYPE>();
    vector <TYPE> * x = new vector<TYPE>();
    vector <TYPE> * y = new vector<TYPE>();
    for (int i=0; i<Nf; i++){
        x->push_back(rand()%100 -50);
        y->push_back(rand()%100-50);
        data->push_back(rand()%100-50);
    }
    for (int i=Nf; i<=Nd; i++){
        y->push_back(rand()%100-50);
        data->push_back(rand()%100-50);
    }
    for (int i=Nd+1; i<=Nd+Nf; i++){
        y->push_back(rand()%100-50);
    }
    // perform forward and adjoint operations
    vector <TYPE>* y_tilde = G(data, x);
    vector <TYPE>* x_tilde = GT(data, y);
    // perform dot product test
    TYPE sumx = 0;
    TYPE * xp = x->data();
    TYPE * xtp = x_tilde->data();
    for (int i=0; i<Nf; i++) sumx += xp[i]*xtp[i];
    TYPE sumy = 0;
    TYPE * yp = y->data();
    TYPE * ytp = y_tilde->data();
    for (int i=0; i<Nd+Nf+1; i++) sumy += yp[i]*ytp[i];
    cout << "sumx = " << sumx << endl;
    cout << "sumy = " << sumy << endl;
}

// Output the L2 norm squared of a vector
TYPE norm2(vector <TYPE>* v){
    TYPE L = 0;
    TYPE * ptr = v->data();
    unsigned int N = v->size();
    for (int i=0; i<N; i++){
        L += ptr[i]*ptr[i];
    }
    return L;
}

// Solve the LSQ problem minimizing r using conjugate gradient method
vector <TYPE>* pef_1d(vector <TYPE>* data, unsigned int Nf, TYPE tolerance){
    //initializing all variables
    unsigned int Nd = data->size()-1;
    TYPE * pdata = data->data();
    // PEF vector
    vector <TYPE>* filk = new vector<TYPE>(Nf);
    TYPE * pfilk = filk->data();
    // Conjugate directions
    vector <TYPE>* cgk_1 = new vector<TYPE>(Nf);
    vector <TYPE>* cgk;
    // intermediate constant
    TYPE betak, betak_1=0;
    // data residuals
    vector <TYPE>* sk = new vector<TYPE>(Nd+Nf+1);
    TYPE * psk = sk->data();
    for (int i=0; i<=Nd; i++) psk[i] = pdata[i];
    // model residuals
    vector <TYPE>* rk = GT(data, sk);
    vector <TYPE>* rk1;
    // objective function
    vector <TYPE> * func = new vector <TYPE>();
    func->push_back(norm2(sk));
    // start the CG loop consisting of 7 steps
    unsigned int k = 0;
    while (k<Nf & (*func)[k]>tolerance){
        // step 1
        TYPE * pcgk_1 = cgk_1->data();
        cgk = new vector<TYPE>(Nf);
        TYPE * pcgk = cgk->data();
        TYPE * prk = rk->data();
        for (int i=0; i<Nf; i++) pcgk[i]= -prk[i] + betak_1*pcgk_1[i];
        // step 2
        vector <TYPE>* denom0k = G(data,cgk);
        vector <TYPE>* denomk = GT(data,denom0k);
        TYPE * pdenomk = denomk->data();
        TYPE sumk = 0;
        for (int i=0; i<Nf; i++) sumk+=pcgk[i]*pdenomk[i];
        TYPE alphak = norm2(rk)/sumk;
        // step 3
        for (int i=0; i<Nf; i++) pfilk[i] += alphak*pcgk[i];
        // step 4
        TYPE * pdenom0k = denom0k->data();
        for (int i=0; i<Nd+Nf+1; i++) psk[i] += + alphak*pdenom0k[i];
        func->push_back(norm2(sk)/(*func)[0]);
        // step 5
        rk1 = GT(data, sk);
        // step 6
        betak_1 = norm2(rk1)/norm2(rk);
        // step 7
        delete cgk_1;
        cgk_1 = cgk;
        delete rk;
        rk = rk1;
        k++;
    }
    delete rk;
    delete cgk_1;
    delete sk;
    // save and plot the objective function
    (*func)[0]=1;
    string PATH = "/Users/nmbader/Desktop/PhD/Cpp/PEF";
    Trace obj(func);
    obj.save(PATH+"/data/functional.txt");
    //plot(PATH+"/data/functional.txt");
    saveplot(PATH+"/data/functional.txt",PATH+"/results/pef_1D_012019/functional","png",
	"set xlabel \"Iteration number\"\n set ylabel \"Normalized residuals L2 norm\"\n");
    // add the leading 1 and return the PEF
    vector<TYPE>::iterator it = filk->begin();
    filk->insert(it,1);
    return filk;
} 


// Estimate LSQ model for Pm = d where P is a PEF
// The data has Nd elements
// The PEF has Nf elements (including the leading 1)
// The residuals can be written as r = d - Pm where
// d = (data 0)' Nd+Nf-1 vector
// m = (m0,...,mNd-1)' the unknown model coefficients
// P is the (Nd+Nf-1)xNd convolution matrix depending solely on the PEF


// Compute the action of P on a given vector v of size=Nd
vector <TYPE>* P(vector <TYPE>* pef, vector <TYPE>* v){
    unsigned int Nf = pef->size();
    unsigned int Nd = v->size();
    vector <TYPE>* out = new vector <TYPE>(Nd+Nf-1);
    TYPE * pf = pef->data();
    TYPE * pd = v->data();
    TYPE * p = out->data();
    for (int i=0; i<Nd; i++){
				for (int j=0; j<Nf; j++){
					p[i+j] = p[i+j] + pd[i]*pf[j];
				}
			}
    return out;
}

// Compute the action of PT (transpose P) on a given vector v of size Nd+Nf-1
vector <TYPE>* PT(vector <TYPE>* pef, vector <TYPE>* v){
    unsigned int Nf = pef->size();
    unsigned int Nd = v->size()-Nf+1;
    vector <TYPE>* out = new vector <TYPE>(Nd);
    TYPE * pf = pef->data();
    TYPE * pd = v->data();
    TYPE * p = out->data();
    for (int i=0; i<Nd; i++){
				for (int j=0; j<Nf; j++){
                    p[i] = p[i] + pd[i+j]*pf[j];
				}
			}
    return out;
}

// Perform the dot product test for P
void P_dot_product(){
    // initialize data, x, and y vectors with random numbers
    unsigned int Nd=1000;
    unsigned int Nf=100;
    vector <TYPE> * pef = new vector<TYPE>();
    vector <TYPE> * x = new vector<TYPE>();
    vector <TYPE> * y = new vector<TYPE>();
    for (int i=0; i<Nf; i++){
        x->push_back(rand()%100 -50);
        y->push_back(rand()%100-50);
        pef->push_back(rand()%100-50);
    }
    for (int i=Nf; i<Nd; i++){
        x->push_back(rand()%100-50);
        y->push_back(rand()%100-50);
    }
    for (int i=Nd+1; i<Nd+Nf; i++){
        y->push_back(rand()%100-50);
    }
    // perform forward and adjoint operations
    vector <TYPE>* y_tilde = P(pef, x);
    vector <TYPE>* x_tilde = PT(pef, y);
    // perform dot product test
    TYPE sumx = 0;
    TYPE * xp = x->data();
    TYPE * xtp = x_tilde->data();
    for (int i=0; i<Nd; i++) sumx += xp[i]*xtp[i];
    TYPE sumy = 0;
    TYPE * yp = y->data();
    TYPE * ytp = y_tilde->data();
    for (int i=0; i<Nd+Nf; i++) sumy += yp[i]*ytp[i];
    cout << "sumx = " << sumx << endl;
    cout << "sumy = " << sumy << endl;
}

// Solve the LSQ problem minimizing r using conjugate gradient method
// This is a deconvolution solver by CGLS
vector <TYPE>* lsq_deconvolution(vector <TYPE>* pef, vector <TYPE>* data, TYPE tolerance){
    //initializing all variables
    unsigned int Nf = pef->size();
    TYPE * ppef = pef->data();
    unsigned int Nd = data->size();
    TYPE * pdata = data->data();
    // Model vector
    vector <TYPE>* mod = new vector<TYPE>(Nd);
    TYPE * pmod = mod->data();
    // Conjugate directions
    vector <TYPE>* cgk_1 = new vector<TYPE>(Nd);
    vector <TYPE>* cgk;
    // intermediate constant
    TYPE betak, betak_1=0;
    // data residuals
    vector <TYPE>* sk = new vector<TYPE>(Nd+Nf-1);
    TYPE * psk = sk->data();
    for (int i=0; i<Nd; i++) psk[i] = -pdata[i];
    // model residuals
    vector <TYPE>* rk = PT(pef, sk);
    vector <TYPE>* rk1;
    // objective function
    vector <TYPE> * func = new vector <TYPE>();
    func->push_back(norm2(sk));
    // start the CG loop consisting of 7 steps
    unsigned int k = 0;
    while (k<Nf & (*func)[k]>tolerance){
        // step 1
        TYPE * pcgk_1 = cgk_1->data();
        cgk = new vector<TYPE>(Nd);
        TYPE * pcgk = cgk->data();
        TYPE * prk = rk->data();
        for (int i=0; i<Nf; i++) pcgk[i]= -prk[i] + betak_1*pcgk_1[i];
        // step 2
        vector <TYPE>* denom0k = P(pef,cgk);
        vector <TYPE>* denomk = PT(pef,denom0k);
        TYPE * pdenomk = denomk->data();
        TYPE sumk = 0;
        for (int i=0; i<Nd; i++) sumk+=pcgk[i]*pdenomk[i];
        TYPE alphak = norm2(rk)/sumk;
        // step 3
        for (int i=0; i<Nd; i++) pmod[i] += alphak*pcgk[i];
        // step 4
        TYPE * pdenom0k = denom0k->data();
        for (int i=0; i<Nd+Nf-1; i++) psk[i] += + alphak*pdenom0k[i];
        func->push_back(norm2(sk)/(*func)[0]);
        // step 5
        rk1 = PT(pef, sk);
        // step 6
        betak_1 = norm2(rk1)/norm2(rk);
        // step 7
        delete cgk_1;
        cgk_1 = cgk;
        delete rk;
        rk = rk1;
        k++;
    }
    delete rk;
    delete cgk_1;
    delete sk;
    // save and plot the objective function
    (*func)[0]=1;
    string PATH = "/Users/nmbader/Desktop/PhD/Cpp/PEF";
    Trace obj(func);
    obj.save(PATH+"/data/functional.txt");
    plot(PATH+"/data/functional.txt");
    return mod;
}